/**
 * (c) Siemens 2011.
 * Licensed under EUPL 1.1
 */
package com.sag.osami.contextstore.serialization.json;

import java.io.Serializable;
import java.util.Map;

import org.codehaus.jackson.map.ObjectMapper;

import com.sag.osami.contextstore.serialization.ISerializationHandler;

/**
 * This serializing utility class uses the <a href="http://jackson.codehaus.org/">Jackson</a> JSON-processor library, which is insanely fast and equally easy to
 * use.
 * <p>
 * Yet there are a few restrictions regarding the classes of the data to be serialized:
 * <ul>
 * <li>all data members must be accessible by common getter/setter methods</li>
 * <li>a default (no-args) constructor must be defined, although it may have any visibility (so <code>private</code> is okay)</li>
 * </ul>
 * Apart from this, there is little which can be done incorrectly: Jackson handles Java's built-in types flawlessly, even when interface types as {@link Map}
 * are used -- then, sensible default types are used during deserialization. Your classes don't have to implement the {@link Serializable} interface. Those
 * kinds of data which are not particularly suited for JSON (e.g. <em>huge</em> <code>byte[]</code> arrays) are handled (space-)efficiently.
 * <p>
 * <em>Please note, that Jackson is not embedded in this bundle but must be present as an own bundle.
 * See the <a href="http://wiki.fasterxml.com/JacksonDownload">Jackson download page</a> for
 * officially packaged OSGi bundles. Both <code>jackson-core</code> and <code>jackson-mapper</code>
 * are necessary.</em>
 * <p>
 * <i>OSGi notice:</i> In order to be able to deserialize all kinds of data, this bundle uses a wildcard <code>Dynamic-Import</code>.
 */
public class JSONSerializer implements ISerializationHandler {

	private static final ObjectMapper json = new ObjectMapper();

	/**
	 * Returns the internally used instance of the Jackson {@link ObjectMapper} for re-use since creation of ObjectMappers can be considered as costly.
	 * 
	 * @return the internally used ObjectMapper instance
	 */
	public static ObjectMapper getJSON() {
		return json;
	}

	public String[] serialize(Object obj) {
		if (obj == null)
			return null;
		try {
			String[] result = new String[] { obj.getClass().getName(), json.writeValueAsString(obj) };
			json.writeValueAsString(""); // clear buffer
			return result;
		} catch (Throwable ex) {
			throw new RuntimeException(ex);
		}
	}

	public Object deserialize(String[] sobj) {
		Object result;
		if (sobj == null)
			return null;
		if (sobj.length != 2)
			throw new IllegalArgumentException("Wrong length of serialized data array: must be 2, but was " + sobj.length);
		try {
			result = json.readValue(sobj[1], Class.forName(sobj[0]));
			json.readValue("[]", Object[].class); // clear buffer
		} catch (Throwable ex) {
			throw new RuntimeException(ex);
		}
		return result;
	}
}