/**
 * (c) Atos IT Solutions and Services GmbH, 2011.
 * Licensed under EUPL 1.1
 */
package com.sag.osami.contextstore.storage.impl;

import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import android.content.ContentValues;
import android.content.Context;
import android.database.Cursor;
import android.database.SQLException;
import android.database.sqlite.SQLiteDatabase;
import android.database.sqlite.SQLiteOpenHelper;

import com.sag.osami.api.context.ContextEvent;
import com.sag.osami.contextstore.storage.DBConstants;
import com.sag.osami.contextstore.storage.SerializedContextStorage;

/**
 * Context store storage implementation based on the Android SQLite API.
 * <p>
 * The table layout is according to {@link DBConstants}. The table is stored in the database file <code>OSAMI</code>.
 * <p>
 * An example query on the context store table would look like this (example fetches last context event with given origin and type):
 * 
 * <pre>
 * public static ContextEvent queryLast(String type, String originID, IContextStore contextStore) {
 * 	String queryGetLastItem = &quot;SELECT * FROM contextstore WHERE type = ? AND originID = ? ORDER BY timestamp LIMIT 1&quot;;
 * 	IContextEvent[] tmp = contextStore.query(queryGetLastItem, type, originID);
 * 	if (tmp.length != 0)
 * 		return tmp[0];
 * 	else
 * 		return null;
 * }
 * </pre>
 * <p>
 * Please note, that all column names (see table layout in {@link DBConstants}) of the ContextStore table must be in the result. While it is most convenient to
 * use <code>"SELECT * FROM contextstore ..."</code> to ensure this, it is not mandatory. The only requirement is, that the result contains at least all column
 * names, therefore allowing complex queries (JOINs etc.) to be performed.
 * <p>
 * Be aware, that storing of context events happens in a background thread, therefore it is not guaranteed that a context event is immediately stored after
 * calling {@link #store(ContextEvent)}.
 * 
 * @see DBConstants
 */
public final class AndroidStorage extends SerializedContextStorage implements DBConstants {

	private static class DBHelper extends SQLiteOpenHelper {

		private static final int DB_VERSION = 1;
		private static final String DB_NAME = "OSAMI";

		public DBHelper(Context context) {
			super(context, DB_NAME, null, DB_VERSION);
		}

		@Override
		public void onCreate(SQLiteDatabase db) {
			StringBuilder createTable = new StringBuilder("CREATE TABLE " + TABLE_NAME + "(");

			// columns: append ", originID TEXT" etc. to SQL string
			String[][] cols = { { COL_ID, "INTEGER PRIMARY KEY" }, { COL_ORIGINID, "TEXT" }, { COL_TYPE, "TEXT" }, { COL_DATA, "TEXT" },
					{ COL_DATA_CLASS, "TEXT" }, { COL_CTXSTORE_ID, "TEXT" }, { COL_TIMESTAMP, "INTEGER" } };
			for (int i = 0; i < cols.length; i++)
				createTable.append((i == 0 ? "" : ", ") + cols[i][0] + ' ' + cols[i][1]);
			createTable.append(")");

			db.execSQL(createTable.toString());
		}

		@Override
		public void onUpgrade(SQLiteDatabase db, int oldVersion, int newVersion) {
			db.execSQL("DROP TABLE IF EXISTS " + TABLE_NAME);
			onCreate(db);
		}

	}

	private DBHelper dbHelper;
	private SQLiteDatabase db;
	private Logger log;
	private ExecutorService dbWorker;
	private Set<Long> deletedIds;
	private long lastID = 0;

	public AndroidStorage(Context context) {
		dbHelper = new DBHelper(context);
		db = dbHelper.getWritableDatabase();
		log = LoggerFactory.getLogger(getClass());
		dbWorker = Executors.newSingleThreadExecutor();
		deletedIds = Collections.synchronizedSet(new TreeSet<Long>());
		queryLastID();
	}

	private void queryLastID() {
		try {
			Cursor c = db.query(TABLE_NAME, new String[] { "MAX(" + COL_ID + ")" }, null, null, null, null, null);
			c.moveToFirst();
			lastID = c.getLong(0);
			c.close();
			log.debug("Last ID in ContextStore DB is " + lastID);
		} catch (Throwable t) {
			// no last ID yet
		}
	}

	@Override
	public void store(final boolean serialize, final ContextEvent context) {
		dbWorker.execute(new Runnable() {
			public void run() {
				try {
					if (deletedIds.remove(context.getID())) // context event with given ID was
															// already marked as to-be-deleted
															// before it had the chance to be
															// actually stored
						return;

					String[] serializedData = (serialize) ? serializer.serialize(context.getData()) : (String[]) context.getData();
					if (serializedData == null)
						serializedData = new String[2];

					ContentValues vals = new ContentValues(7);
					vals.put(COL_ID, context.getID());
					vals.put(COL_ORIGINID, context.getOriginID());
					vals.put(COL_TYPE, context.getType());
					vals.put(COL_DATA_CLASS, serializedData[0]);
					vals.put(COL_DATA, serializedData[1]);
					vals.put(COL_TIMESTAMP, context.getTimestamp().getTime());
					vals.put(COL_CTXSTORE_ID, context.getContextStoreOriginID());

					db.insertOrThrow(TABLE_NAME, null, vals);
				} catch (SQLException e) {
					log.error("Persisting of ContextEvent failed!", e);
				}
			}
		});
	}

	@Override
	public void delete(Long contextId) {
		int deleted = db.delete(TABLE_NAME, COL_ID + " = ?", new String[] { contextId.toString() });
		if (deleted == 0)
			deletedIds.add(contextId);
	}

	@Override
	public void delete(List<Long> contextIds) {
		for (Long contextId : contextIds)
			delete(contextId);
	}

	@Override
	public void delete(String query, Object... params) {
		db.execSQL(query, params);
	}

	@Override
	public Long retrieveLastID() {
		return lastID;
	}

	@Override
	public Iterable<ContextEvent> queryLazy(final boolean deserialize, final String query, final Object... params) throws UnsupportedOperationException {
		return new Iterable<ContextEvent>() {
			@Override
			public Iterator<ContextEvent> iterator() {
				return new IteratorImpl(deserialize, query, params);
			}
		};
	}

	@Override
	public ContextEvent[] query(boolean deserialize, String query, Object... params) {
		return IteratorUtil.toArray(new IteratorImpl(deserialize, query, params));
	}

	private static String loadQuery(List<Long> contextIds) {
		StringBuilder idParams = new StringBuilder(contextIds.size() * 2 - 1);
		for (int i = 0; i < contextIds.size(); i++)
			idParams.append(i == 0 ? "?" : ",?");
		return "SELECT * FROM " + TABLE_NAME + " WHERE " + COL_ID + " IN (" + idParams + ")";
	}

	@Override
	public Iterable<ContextEvent> loadLazy(boolean deserialize, List<Long> contextIds) throws UnsupportedOperationException {
		return queryLazy(deserialize, loadQuery(contextIds), contextIds.toArray());
	}

	@Override
	public ContextEvent[] load(boolean deserialize, List<Long> contextIds) {
		return query(deserialize, loadQuery(contextIds), contextIds.toArray());
	}

	@Override
	public ContextEvent load(boolean deserialize, Long contextId) {
		ContextEvent[] loaded = load(deserialize, Arrays.asList(contextId));
		return (loaded.length == 0) ? null : loaded[0];
	}

	private static Map<String, Integer> columnIndexesForContextEvents(Cursor cur) {
		HashMap<String, Integer> cols = new HashMap<String, Integer>();
		for (String col : COLS)
			cols.put(col, cur.getColumnIndexOrThrow(col));
		return cols;
	}

	private static String[] formatParams(Object[] params) {
		String[] strparams = new String[params.length];
		for (int i = 0; i < params.length; i++) {
			if (params[i] instanceof Date)
				strparams[i] = String.valueOf(((Date) params[i]).getTime());
			else
				strparams[i] = params[i].toString();
		}
		return strparams;
	}

	@Override
	protected void finalize() throws Throwable {
		db.close();
		super.finalize();
	}

	private class IteratorImpl extends IteratorUtil.RowIterator {

		private Cursor cur;
		private Map<String, Integer> cols;

		protected IteratorImpl(boolean deserialize, String query, Object[] params) {
			super(AndroidStorage.this.log, deserialize, query, params);
		}

		@Override
		protected void initialize() throws IteratorException {
			try {
				cur = db.rawQuery(query, formatParams(params));
				cols = columnIndexesForContextEvents(cur);
			} catch (SQLException e) {
				throw new IteratorException(e);
			}
		}

		@Override
		protected boolean moveNext() throws IteratorException {
			return cur.moveToNext();
		}

		@Override
		protected ContextEvent loadContextEvent() throws IteratorException {
			try {
				Long id = cur.getLong(cols.get(COL_ID));
				String originID = cur.getString(cols.get(COL_ORIGINID));
				String ctxstoreID = cur.getString(cols.get(COL_CTXSTORE_ID));
				String type = cur.getString(cols.get(COL_TYPE));
				Date timestamp = new Date(cur.getLong(cols.get(COL_TIMESTAMP)));
				String[] sdata = new String[] { cur.getString(cols.get(COL_DATA_CLASS)), cur.getString(cols.get(COL_DATA)) };
				Object data = null;

				// there *is* payload data
				if (sdata[0] != null)
					data = (deserialize) ? serializer.deserialize(sdata) : sdata;

				return new ContextEvent(id, data, originID, timestamp, type, ctxstoreID);
			} catch (SQLException e) {
				throw new IteratorException(e);
			}
		}

		@Override
		protected void close() {
			cur.close();
		}

		@Override
		protected void finalize() throws Throwable {
			close();
			super.finalize();
		}

	}
}
