/**
 * (c) Atos IT Solutions and Services GmbH, 2011.
 * Licensed under EUPL 1.1
 */
package com.sag.osami.contextstore.storage.jdbc;

import com.sag.osami.contextstore.ContextStore;
import com.sag.osami.contextstore.storage.DBConstants;
import com.sag.osami.contextstore.storage.impl.JDBCStorage;

/**
 * This interface provides methods for getting SQL statements for the various tasks a storage implementation must handle, i.e.:
 * <ul>
 * <li>inserting</li>
 * <li>loading</li>
 * <li>deleting</li>
 * </ul>
 * of ContextEvents (by ID(s)) as well as statements for creating and dropping the ContextStore table and also for querying the last given ContextEvent ID
 * [before the ContextStore was shutdown].
 * <p>
 * This interface abstracts away the various small SQL syntax differences between database server implementations and enables to re-use the same JDBC-based
 * storage implementation for all JDBC-compatible DBs.
 * <p>
 * An implementation of a JDBC dialect is typically extremely lightweight as it simply must return a bunch of (usually) static SQL strings.
 * <p>
 * <b>Important note:</b> Most of the query strings are used in a context where parameter binding is performed; those parameters are in a well-defined order and
 * the query strings should contain the according "?" placeholders. See the query methods documentation for details.
 * 
 * @see JDBCStorage
 * @see GenericJDBCDialect
 */
public interface IJDBCDialect extends DBConstants {

	/**
	 * Returns the fully qualified class name of the JDBC driver to be used.
	 * 
	 * @return the JDBC driver class name
	 */
	String driver();

	/**
	 * Returns the SQL statement for querying the highest/last ID present in the context store table.
	 * 
	 * @return the last ID SQL statement
	 */
	String queryLastID();

	// DDL statements

	/**
	 * Returns the SQL statement for checking whether the context store table already exists.
	 * 
	 * @return the check SQL statement
	 */
	String tableExists();

	/**
	 * Returns the SQL statement for creating the context store table.
	 * <p>
	 * Be aware that the ID column should not be auto incremented by the database. ID generation is handled by the {@link ContextStore}.
	 * 
	 * @return the create table SQL statement
	 */
	String createTable();

	/**
	 * Returns the SQL statement for dropping the context store table.
	 * 
	 * @return the drop table SQL statement
	 */
	String dropTable();

	// ContextEvent inserting/loading/deleting
	/**
	 * Returns the SQL statement for inserting a context event. It contains parameters (?) for each column, ordered as specified in {@link DBConstants#COLS}.
	 * 
	 * @return the insert SQL statement
	 */
	String insertContextEvent();

	/**
	 * Returns the SQL statement for loading several context events. It contains <code>&lt;idCount&gt;</code> parameters (?) which are bound to the context
	 * event IDs.
	 * <p>
	 * This operation is optional. If it is not supported, {@link JDBCStorage} will load the context events individually, querying one context event at a time.
	 * 
	 * @param idsCount
	 *            number of the context event IDs to be loaded
	 * @return the load SQL statement
	 * @throws UnsupportedOperationException
	 *             if the underlying JDBC database does not support loading multiple context events in one query
	 */
	String loadContextEvents(int idsCount) throws UnsupportedOperationException;

	/**
	 * Returns the SQL statement for loading a context event. It contains one parameter (?) which is bound to the context event ID.
	 * 
	 * @return the load SQL statement
	 */
	String loadContextEvent();

	/**
	 * Returns the SQL statement for deleting a context event. It contains one parameter (?) which is bound to the context event ID.
	 * 
	 * @return the delete SQL statement
	 */
	String deleteContextEvent();

}